""" fprime_gds.executables.dictionary_merge: script to merge two F Prime dictionaries """

import argparse
import functools
import json
import re
import sys
from pathlib import Path


def validate_metadata(metadata1, metadata2):
    """ Check consistency between metadata blocks
    
    The JSON dictionary has multiple fields in the metadata block. This function will check that there is consistency
    between these two blocks.

    Args:
        metadata1: metadata from the first dictionary
        metadata2: metadata from the second dictionary
    """
    for field in ["projectVersion", "frameworkVersion", "dictionarySpecVersion"]:
        value1 = metadata1[field]
        value2 = metadata2[field]
        if value1 != value2:
            raise ValueError(f"Inconsistent metadata values for field '{field}'. ({value1} vs {value2})")

def validate_non_unique(non_unique1, non_unique2):
    """ Validate non-unique definitions are consistent between dictionaries """
    indexed_non_unique1 = {value.get("qualifiedName"): value for value in non_unique1}

    for value2 in non_unique2:
        value1 = indexed_non_unique1.get(value2["qualifiedName"], None)
        if value1 is not None and value1 != value2:
            raise ValueError(f"'{value2['qualifiedName']}' has inconsistent definitions")

def validate_unique(unique1, unique2):
    """ Validate unique definitions have no duplication """
    ids = {item.get("id", item.get("opcode", "")) for item in unique1}
    names = {item.get("name") for item in unique1}


    for value2 in unique2:
        name = value2['name']
        id = value2.get("id", value2.get("opcode", ""))
        if name in names:
            raise ValueError(f"'{name}' appears in both dictionaries")
        if id and id in ids:
            raise ValueError(f"ID/Opcode {id} used in both dictionaries")


def merge_metadata(meta1, meta2, name=None, permissive=False):
    """ Merge JSON dictionary metadata blocks
    
    The JSON dictionary starts with a metadata block. This function will merge the two metadata blocks preferring the
    first when there is a discrepancy. 'name' will be supplied as the new name defaulting to "name1_name2_merged" when
    not supplied. If 'permissive' is true, version discrepancies will be ignored otherwise this will throw a ValueError
    if the versions do not match.

    Args:
        meta1: first metadata block
        meta2: second metadata block
        name: (optional) name for the new dictionary (Default: meta.name_meta2.name_merged)
        permissive: (optional) True to allow version miss-matching. (Default: False)
    Return:
        merged metadata block
    Throws:
        ValueError on version miss-match without the permissive flag
    """
    if not permissive:
        validate_metadata(meta1, meta2)
    if name is None:
        name = f"{meta1.get('deploymentName', 'unknown')}_{meta2.get('deploymentName', 'unknown')}_merged"
    return {
        **meta1,
        **{
            "deploymentName": name
          }
    }

def merge_lists(list1, list2, validator):
    """ Merge list-like entities
    
    This will merge two list-like entities using the supplied validator.

    Args:
        list1: first list-like
        list2: second list-like
        validator: validate the lists are consistent or non-colliding

    """
    validator(list1, list2)
    singular = {item.get("qualifiedName", item.get("name", "")): item for item in list1 + list2}
    return list(singular.values())

def merge_non_unique(non_unique1, non_unique2):
    """ Merge the non-unique blocks in JSON dictionaries

    JSON dictionaries have some non-unique definitions (e.g. "typeDefinitions") that must be merged ensuring
    consistency but ignoring duplication. This function will create a superset of the two blocks. Inconsistent
    definitions will result in a ValueError.

    Args:
        non_unique1: first non unique block
        non_unique2: second non unique block
    """
    return merge_lists(non_unique1, non_unique2, validate_non_unique)


def merge_unique(unique1, unique2):
    """ Merge the unique blocks in JSON dictionaries

    JSON dictionaries have some unique definitions (e.g. "eventDefinitions") that must be merged ensuring that entries
    are not duplicated between the sets. This function will create a superset of the two blocks. Duplicated definitions
    will result in a ValueError.

    Args:
        unique1: first unique block
        unique2: second unique block
    """
    return merge_lists(unique1, unique2, validate_unique)


def merge_dictionaries(dictionary1, dictionary2, name=None, permissive=False):
    """ Merge two dictionaries

    This will merge two JSON dictionaries' major top-level sections. Unknown fields will be preserved preferring
    dictionary1's content for unknown fields.

    Args:
        dictionary1: dictionary 1's content
        dictionary2: dictionary 2's content
        name: new 'deploymentName' field
        permissive: allow miss-matched dictionary versions

    Return: merged dictionaries
        
    """
    merge_metadata_fn = functools.partial(merge_metadata, name=name, permissive=permissive)

    stages = [
        ("metadata", merge_metadata_fn),
        ("typeDefinitions", merge_non_unique),
        ("constants", merge_non_unique),
        ("commands", merge_unique),
        ("parameters", merge_unique),
        ("events", merge_unique),
        ("telemetryChannels", merge_unique),
        ("records", merge_unique),
        ("containers", merge_unique),
        ("telemetryPacketSets", merge_unique),

    ]

    merged = {**dictionary2, **dictionary1}
    for field, merger in stages:
        object1 = dictionary1[field]
        object2 = dictionary2[field]
        try:
            merged[field] = merger(object1, object2)
        except ValueError as value_error:
            raise ValueError(f"Merging '{field}' failed. {value_error}")
        except KeyError as key_error:
            raise ValueError(f"Malformed dictionary section '{field}'. Missing key: {key_error}")
    return merged

def parse_arguments():
    """ Parse arguments for this script """
    parser = argparse.ArgumentParser(description="Merge two dictionaries")
    parser.add_argument("--name", type=str, default=None, help="Name to use as the new 'deploymentName' field")
    parser.add_argument("--output", type=Path, default=Path("MergedAppDictionary.json"),
                        help="Output dictionary path. Default: MergedAppDictionary.json")
    parser.add_argument("--permissive", action="store_true", default=False,
                        help="Ignore discrepancies between dictionaries")
    parser.add_argument("dictionary1", type=Path, help="Primary dictionary to merge")
    parser.add_argument("dictionary2", type=Path, help="Secondary dictionary to merge")

    args = parser.parse_args()

    # Validate arguments
    if args.name is not None and not re.match("[a-zA-Z_][a-zA-Z_0-9]*"):
        raise ValueError(f"--name '{args.name}' is an invalid identifier")
    if not args.dictionary1.exists():
        raise ValueError(f"'{args.dictionary1}' does not exist")
    if not args.dictionary2.exists():
        raise ValueError(f"'{args.dictionary2}' does not exist")
    return args

def main():
    """ Main entry point """
    try:
        args = parse_arguments()
        # Open dictionaries
        with open(args.dictionary1, "r") as dictionary1_fh:
            dictionary1 = json.load(dictionary1_fh)
        with open(args.dictionary2, "r") as dictionary2_fh:
            dictionary2 = json.load(dictionary2_fh)
        output = merge_dictionaries(dictionary1, dictionary2, args.name, args.permissive)
        with open(args.output, "w") as output_fh:
            json.dump(output, output_fh, indent=2)
    except Exception as exception:
        print(f"[ERROR] {exception}", file=sys.stderr)
        sys.exit(1)
    sys.exit(0)

if __name__ == "__main__":
    main()



import os
import sys
import unittest

from fprime.common.models.serialize.numerical_types import I32Type
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.time_type import TimeType
from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.event_template import EventTemplate
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.utils.event_severity import EventSeverity

filename = os.path.dirname(__file__)
gdsName = os.path.join(filename, "../../../../src")
fprimeName = os.path.join(filename, "../../../../../Fw/Python/src")
sys.path.insert(0, gdsName)
sys.path.insert(0, fprimeName)


class PredicateTestCases(unittest.TestCase):
    @staticmethod
    def check_str(pred):
        """
        Tests for the predicate class and it's functionality.
        """
        try:
            pred.__str__()
            print(pred)
            assert True, f"predicate provides string summary: {str(pred)}"
        except NotImplementedError:
            assert False, "invoking str(pred) was not supported"

    @staticmethod
    def test_NotImplemented():
        class dummyPred(predicates.predicate):
            pass

        pred = dummyPred()

        try:
            pred.__call__(2)
            assert (
                False
            ), "invoking an incomplete subclass didn't raise NotImplementedError"
        except NotImplementedError:
            assert True, "invoking an incomplete subclass raised NotImplementedError"

        try:
            str(pred)
            assert (
                False
            ), "invoking an incomplete subclass didn't raise NotImplementedError"
        except NotImplementedError:
            assert True, "invoking an incomplete subclass raised NotImplementedError"

    @staticmethod
    def test_Implemented():
        pred = predicates.less_than(2)
        try:
            pred.__call__(2)
        except NotImplementedError:
            assert (
                False
            ), "invoking __call__ on an complete subclass of predicate failed"

        try:
            str(pred)
        except NotImplementedError:
            assert False, "invoking __str__ on an complete subclass of predicate failed"
        assert True, "implemented predicate had no problems invoking functions"

    @staticmethod
    def test_is_predicate():
        class dummyPred:
            def __call__(self, item):
                pass

            def __str__(self):
                return "dummyPred"

        pred = dummyPred()
        assert predicates.is_predicate(
            pred
        ), "a class with call and str methods satisfies is_predicate()"
        pred = object()
        assert not predicates.is_predicate(
            pred
        ), "a class without call and str doesn't satisfy is_predicate()"
        pred = predicates.predicate()
        assert predicates.is_predicate(
            pred
        ), "an instance of predicate satisfies is_predicate()"
        pred = predicates.equal_to(1)
        assert predicates.is_predicate(
            pred
        ), "an instance of a subclass of predicate satisfies is_predicate()"

    def test_less_than(self):
        pred = predicates.less_than(2)
        assert pred(1), "one is less than two"
        assert pred(-1), "negative one is less than two"
        assert not pred(2), "two is not less than two"
        assert not pred(3), "three is not less than two"
        self.check_str(pred)

    def test_greater_than(self):
        pred = predicates.greater_than(2)
        assert pred(3), "three is greater than two"
        assert pred(100), "one hundred is greater than two"
        assert not pred(2), "two is not greater than two"
        assert not pred(-1), "negative one is not greater than two"
        self.check_str(pred)

    def test_equal_to(self):
        pred = predicates.equal_to(1)
        assert pred(1), "one is equal to one"
        assert not pred(100), "one hundred is not equal to one"
        assert not pred(2), "two is not equal to one"
        assert not pred(-1), "negative one is not equal to one"
        self.check_str(pred)

    def test_not_equal_to(self):
        pred = predicates.not_equal_to(1)
        assert pred(0), "zero is not equal to one"
        assert pred(-1), "negative one is not equal to one"
        assert pred(2), "two is not equal to one"
        assert not pred(1), "one is not not equal to one"
        self.check_str(pred)

    def test_less_than_or_equal_to(self):
        pred = predicates.less_than_or_equal_to(1)
        assert pred(1), "one is less than or equal to 1"
        assert pred(-1), "negative one is less than or equal to 1"
        assert not pred(10), "ten is not less than or equal to 1"
        self.check_str(pred)

    def test_greater_than_or_equal_to(self):
        pred = predicates.greater_than_or_equal_to(1)
        assert pred(1), "one is greater than or equal to 1"
        assert pred(10), "ten is greater than or equal to 1"
        assert not pred(-1), "negative is not greater than or equal to 1"
        self.check_str(pred)

    def test_within_range(self):
        pred = predicates.within_range(5, 10)
        assert pred(5), "5 is the lower bound and within [5,10]"
        assert pred(10), "10 is the upper bound and within [5,10]"
        assert pred(7.5), "7.5 is within [5,10]"
        assert not pred(2), "2 is not within [5,10]"
        assert not pred(11), "2 is not within [5,10]"
        self.check_str(pred)

    def test_is_a_member_of(self):
        pred = predicates.is_a_member_of(["a", "b", "c"])
        assert pred("a"), "a is in the set [a, b, c]"
        assert pred("b"), "b is in the set [a, b, c]"
        assert pred("c"), "c is in the set [a, b, c]"
        assert not pred(1), "1 is not in the set [a, b, c]"
        assert not pred("x"), "x is not in the set [a, b, c]"
        self.check_str(pred)

    def test_is_not_a_member_of(self):
        pred = predicates.is_not_a_member_of(["a", "b", "c"])
        assert pred(1), "1 is not in the set [a, b, c]"
        assert pred("x"), "x is not in the set [a, b, c]"
        assert not pred("a"), "a is not not in the set [a, b, c]"
        assert not pred("b"), "b is not not in the set [a, b, c]"
        assert not pred("c"), "c is not not in the set [a, b, c]"
        self.check_str(pred)

    def test_always_true(self):
        pred = predicates.always_true()
        assert pred(1), "numbers are true"
        assert pred("string"), "strings are true"
        assert pred(object()), "an object is true"
        self.check_str(pred)

    def test_invert(self):
        pred = predicates.invert(predicates.always_true())
        assert not pred(1), "numbers are not true"
        assert not pred("string"), "strings are not true"
        assert not pred(object()), "an object is not true"
        self.check_str(pred)

    def test_satisfies_all(self):
        p_list = [predicates.less_than(8)]
        p_list.append(predicates.less_than(6))
        p_list.append(predicates.equal_to(4))
        pred = predicates.satisfies_all(p_list)
        assert pred(4), "4 satisfies all predicates in the list"
        assert not pred(5), "5 satisfies only 2 predicates in the list"
        assert not pred(7), "7 satisfies only 1 predicate in the list"
        assert not pred(9), "9 satisfies only no predicates in the list"
        self.check_str(pred)

    def test_satisfies_any(self):
        p_list = [predicates.less_than(8)]
        p_list.append(predicates.less_than(6))
        p_list.append(predicates.equal_to(4))
        pred = predicates.satisfies_any(p_list)
        assert pred(4), "4 satisfies all predicates in the list"
        assert pred(5), "5 satisfies only 2 predicates in the list"
        assert pred(7), "7 satisfies only 1 predicate in the list"
        assert not pred(9), "9 satisfies only no predicates in the list"
        self.check_str(pred)

    @staticmethod
    def test_args_predicates():
        a_list = ["a", "p", "p", "l", "e"]
        pred = predicates.args_predicate(["a", "p", "p", "l", "e"])
        assert pred(a_list), f"The list {a_list} should have been accepted"
        a_list[4] = "r"
        assert not pred(a_list), f"The list {a_list} should not have been accepted"
        a_list = ["a", "p", "p", "l"]
        assert not pred(a_list), f"The list {a_list} should not have been accepted"

        a_list = ["a", "p", "p", "l", "e"]
        pred = predicates.args_predicate(["a", "p", "p", "l", None])
        assert pred(a_list), f"The list {a_list} should have been accepted"
        a_list[4] = 7
        assert pred(a_list), f"The list {a_list} should have been accepted"
        a_list[4] = "r"
        assert pred(a_list), f"The list {a_list} should have been accepted"

        l_pred = predicates.within_range(0, 10)
        pred = predicates.args_predicate([l_pred, 2, 3, 4, 5, 6])

        n_list = [1, 2, 3, 4, 5, 6]
        assert pred(n_list), f"The list {n_list} should have been accepted"

        for i in range(10):
            n_list[0] = i
            assert pred(n_list), f"The list {n_list} should have been accepted"
        n_list[0] = -5
        assert not pred(n_list), f"The list {n_list} should not have been accepted"
        n_list[0] = 15
        assert not pred(n_list), f"The list {n_list} should not have been accepted"

        pred = predicates.args_predicate(8)
        assert pred(8), "The value 8 should have been accepted."

    def test_telemetry_predicates(self):
        test_string_type = StringType.construct_type("TestCh2String")
        temp1 = ChTemplate(1, "Test Channel 1", "Predicate_Tester", I32Type)
        temp2 = ChTemplate(2, "Test Channel 2", "Predicate_Tester", test_string_type)
        update1 = ChData(I32Type(20), TimeType(), temp1)
        update2 = ChData(test_string_type("apple"), TimeType(), temp2)

        pred = predicates.telemetry_predicate()
        assert pred(
            update1
        ), "If no fields are specified a ChData object should return True"
        assert pred(
            update2
        ), "If no fields are specified a ChData object should return True"
        assert not pred(
            "diff object"
        ), "Anything that's not a ChData object should be False"
        assert not pred(5), "Anything that's not a ChData object should be False"
        self.check_str(pred)

        id_pred = predicates.equal_to(1)
        pred = predicates.telemetry_predicate(id_pred=id_pred)
        assert pred(update1), "This predicate on the ID 1 should return True"
        assert not pred(update2), "This predicate on the ID 2 should return False"
        self.check_str(pred)

        val_pred = predicates.equal_to("apple")
        pred = predicates.telemetry_predicate(value_pred=val_pred)
        assert not pred(update1), "This predicate on the value 20 should return False"
        assert pred(update2), 'This predicate on the value "apple" should return True'
        self.check_str(pred)

        time_pred = predicates.equal_to(0)
        pred = predicates.telemetry_predicate(time_pred=time_pred)
        assert pred(update1), "This predicate on the time 0 should return True"
        assert pred(update2), "This predicate on the time 0 should return True"
        self.check_str(pred)

        val_pred = predicates.within_range(10, 30)
        pred = predicates.telemetry_predicate(id_pred, val_pred, time_pred)
        assert pred(update1), "Specifying all fields should return True for update 1"
        assert not pred(
            update2
        ), "Specifying all fields should return False for update 2"
        self.check_str(pred)

    def test_event_predicates(self):
        test_string_type = StringType.construct_type("TestEventString")
        args1_def = [("name", "string", test_string_type), ("age", "int", I32Type)]
        temp1 = EventTemplate(
            1,
            "Test Msg 1",
            "Predicate Tester",
            args1_def,
            EventSeverity.ACTIVITY_LO,
            "",
        )
        args1 = (test_string_type("John"), I32Type(35))
        msg1 = EventData(args1, TimeType(), temp1)
        args2_def = [
            ("description", "string", test_string_type),
            ("count", "int", I32Type),
        ]
        temp2 = EventTemplate(
            2,
            "Test Msg 2",
            "Predicate Tester",
            args2_def,
            EventSeverity.ACTIVITY_HI,
            "",
        )
        args2 = (test_string_type("Dozen"), I32Type(12))
        msg2 = EventData(args2, TimeType(), temp2)

        pred = predicates.event_predicate()
        assert pred(
            msg1
        ), "If no fields are specified an EventData object should return True"
        assert pred(
            msg2
        ), "If no fields are specified an EventData object should return True"
        assert not pred(
            "diff object"
        ), "Anything that's not an EventData object should be False"
        assert not pred(5), "Anything that's not a EventData object should be False"
        self.check_str(pred)

        id_pred = predicates.equal_to(1)
        pred = predicates.event_predicate(id_pred=id_pred)
        assert pred(msg1), "This predicate on the ID 1 should return True"
        assert not pred(msg2), "This predicate on the ID 2 should return False"
        self.check_str(pred)

        args_pred = predicates.args_predicate([None, None])
        pred = predicates.event_predicate(args_pred=args_pred)
        assert pred(
            msg1
        ), "This predicate should return True, as it expects an event with 2 args"
        assert pred(
            msg2
        ), "This predicate should return True, as it expects an event with 2 args"
        self.check_str(pred)

        args_pred = predicates.args_predicate(["John", 35])
        pred = predicates.event_predicate(args_pred=args_pred)
        assert pred(
            msg1
        ), "This predicate should return True as msg1 has args (str John, int32 35)"
        assert not pred(
            msg2
        ), "This predicate should return False as msg2 has args (str Dozen, int32 12)"
        self.check_str(pred)

        severity_pred = predicates.equal_to(EventSeverity.ACTIVITY_LO)
        pred = predicates.event_predicate(severity_pred=severity_pred)
        assert severity_pred(msg1.get_severity())
        assert pred(
            msg1
        ), "This predicate should return True as msg1 has an ACTIVITY_LO severity"
        assert not pred(
            msg2
        ), "This predicate should return False as msg2 has an ACTIVITY_HI severity"
        self.check_str(pred)

        time_pred = predicates.equal_to(0)
        pred = predicates.event_predicate(time_pred=time_pred)
        assert pred(msg1), "This predicate on the time 0 should return True"
        assert pred(msg2), "This predicate on the time 0 should return True"
        self.check_str(pred)

        pred = predicates.event_predicate(id_pred, args_pred, severity_pred, time_pred)
        assert pred(msg1), "Specifying all fields should return True for msg1"
        assert not pred(msg2), "Specifying all fields should return False for msg2"
        self.check_str(pred)


if __name__ == "__main__":
    unittest.main()



"""
@brief Channel Decoder class used to parse binary channel telemetry data

Decoders are responsible for taking in serialized data and parsing it into
objects. Decoders receive serialized data (that had a specific descriptor) from
a distributer that it has been registered to. The distributer will send the
binary data after removing any length and descriptor headers.

Example data that would be sent to a decoder that parses channels:
    +-------------------+---------------------+------------ - - -
    | ID (4 bytes) | Time Tag (11 bytes) | Data....
    +-------------------+---------------------+------------ - - -

@date Created July 11, 2018
@author R. Joseph Paetz

@bug No known bugs
"""
from fprime.common.models.serialize.time_type import TimeType

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.decoders.decoder import Decoder, DecodingException
from fprime_gds.common.utils import config_manager


class ChDecoder(Decoder):
    """Decoder class for Channel data"""

    def __init__(self, ch_dict, config):
        """
        ChDecoder class constructor

        Args:
            ch_dict: Channel telemetry dictionary. Channel IDs should be keys
                     and ChTemplate objects should be values

        Returns:
            An initialized channel decoder object.
        """
        super().__init__()

        if config is None:
            # Retrieve singleton for the configs
            config = config_manager.ConfigManager.get_instance()

        self.__dict = ch_dict
        self.id_obj = config.get_type("FwChanIdType")

    def decode_api(self, data):
        """
        Decodes the given data and returns the result.

        This function allows for non-registered code to call the same decoding
        code as is used to parse data passed to the data_callback function.

        Args:
            data: Binary telemetry channel data to decode

        Returns:
            Parsed version of the channel telemetry data in the form of a
            ChData object or None if the data is not decodable
        """
        ptr = 0

        # list for decoded channel values
        ch_list = []

        while (ptr < len(data)):

            # Decode Ch ID here...
            self.id_obj.deserialize(data, ptr)
            ptr += self.id_obj.getSize()
            ch_id = self.id_obj.val

            # Decode time...
            ch_time = TimeType()
            ch_time.deserialize(data, ptr)
            ptr += ch_time.getSize()

            if ch_id not in self.__dict:
                msg = f"Channel {ch_id} not found in dictionary"
                raise DecodingException(msg)
            # Retrieve the template instance for this channel
            ch_temp = self.__dict[ch_id]

            try:
                val_obj = self.decode_ch_val(data, ptr, ch_temp)
            except Exception as exc:
                msg = f"Channel {ch_temp.name} failed to decode: {exc}"
                raise DecodingException(msg)
            ch_list.append(ChData(val_obj, ch_time, ch_temp))
            ptr += val_obj.getSize()
        return ch_list

    @staticmethod
    def decode_ch_val(val_data, offset, template):
        """
        Decodes the given channel's value from the given data

        Args:
            val_data: Data to parse the value out of
            offset: Location in val_data to start the parsing
            template: Channel Template object for the channel

        Returns:
            The channel's value as an instance of a class derived from
            the BaseType class. The val_data has been deserialized using this
            object, and so the channel value can be retrieved from the obj's
            val field.
        """
        val_obj = template.get_type_obj()()
        val_obj.deserialize(val_data, offset)
        return val_obj



import math
import os
import threading
import time
import unittest

# these imports are needed to generate data objects.
from fprime.common.models.serialize.numerical_types import I32Type, U32Type
from fprime.common.models.serialize.time_type import TimeType
from fprime_gds.common.models.dictionaries import Dictionaries
from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.cmd_data import CmdData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.history.test import TestHistory
from fprime_gds.common.pipeline.standard import StandardPipeline
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI
from fprime_gds.common.utils.config_manager import ConfigManager


class UTPipeline(StandardPipeline):
    """
    This pipeline shares many of the same calls available in pipeline.standard. It
    is used by this testcase to feed simulated data to the test api.
    """

    def __init__(self):
        self.command_count = 0
        self.t0 = TimeType()
        StandardPipeline.__init__(self)

    def connect(self, address, port):
        pass

    def disconnect(self):
        # Standard pipeline starts uplink thread that must be stopped
        self.files.uplinker.exit()

    def send_command(self, command, args):
        command_template = self.dictionaries.command_id[command]
        cmd_data = CmdData(tuple(args), command_template)
        self.histories.commands.data_callback(cmd_data)
        for hist in self.coders.command_subscribers:
            hist.data_callback(cmd_data)

        ev_temp = self.dictionaries.event_name["apiTester.CommandReceived"]
        event = EventData((U32Type(cmd_data.get_id()),), self.t0 + time.time(), ev_temp)
        self.enqueue_event(event)

        ev_temp = self.dictionaries.event_name["apiTester.HistorySizeUpdate"]
        evr_size = U32Type(len(self.histories.events.retrieve()))
        cmd_size = U32Type(len(self.histories.commands.retrieve()))
        ch_size = U32Type(len(self.histories.channels.retrieve()))
        event = EventData((evr_size, cmd_size, ch_size), self.t0 + time.time(), ev_temp)
        self.enqueue_event(event)

        self.command_count += 1
        ch_temp = self.dictionaries.channel_name["apiTester.CommandCounter"]
        update = ChData(U32Type(self.command_count), self.t0 + time.time(), ch_temp)
        self.enqueue_telemetry(update)

    def enqueue_event(self, event):
        """
        Used by the unit test to feed simulated data objects into the pipeline
        """
        self.coders.event_decoder.send_to_all(event)

    def enqueue_telemetry(self, channel):
        """
        Used by the unit test to feed simulated data objects into the pipeline
        """
        self.coders.channel_decoder.send_to_all(channel)


class APITestCases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pipeline = UTPipeline()
        config = ConfigManager.get_instance()
        path = os.path.join(os.path.dirname(__file__), "./UnitTestDictionary.xml")

        # Load test dictionary
        dictionaries = Dictionaries()
        dictionaries.load_dictionaries(
            path, None, None
        )

        file_store = os.path.join(os.path.dirname(__file__), "./")
        cls.pipeline.setup(config, dictionaries, file_store)
        log_path = os.path.join(os.path.dirname(__file__), "./logs")
        cls.api = IntegrationTestAPI(cls.pipeline, log_path)
        cls.case_list = []  # TODO find a better way to do this.
        cls.threads = []

    def setUp(self):
        for t in self.threads:
            if t.is_alive():
                t.join()
        self.threads.clear()
        count = len(self.case_list)
        self.api.start_test_case(self._testMethodName, count)
        self.case_list.append(1)
        self.tHistory = TestHistory()
        self.t0 = TimeType()

    @classmethod
    def tearDownClass(cls):
        cls.pipeline.disconnect()
        cls.api.teardown()

    ######################################################################################
    #   Test Case Helper Methods
    ######################################################################################
    def fill_history(self, callback, items, timestep=0.0):
        for item in items:
            if timestep:
                time.sleep(timestep)
            if isinstance(item, (ChData, EventData)) and item.time == 0:
                item.time = self.t0 + time.time()
            callback(item)

    def fill_history_async(self, callback, items, timestep=1.0):
        t = threading.Thread(target=self.fill_history, name="FillHistoryAsync", args=(callback, items, timestep))
        self.threads.append(t)
        t.start()
        return t

    @staticmethod
    def assert_lists_equal(expected, actual):
        assert len(expected) == len(
            actual
        ), f"the given list should have had the length {len(expected)}, but instead had {len(actual)}\nExpected {expected}\nActual{actual}"
        for i, item in enumerate(expected):
            assert (
                item == actual[i]
            ), f"the {i} element of the expected list should be {item}, but was {actual[i]}."

    def get_counter_sequence(self, length):
        seq = []
        for i in range(length):
            ch_temp = self.pipeline.dictionaries.channel_name["apiTester.Counter"]
            seq.append(ChData(U32Type(i), TimeType(), ch_temp))
        return seq

    def get_oscillator_sequence(self, length):
        seq = []
        for i in range(length):
            ch_temp = self.pipeline.dictionaries.channel_name["apiTester.Oscillator"]
            val = int(round(10 * math.sin(math.radians(i))))
            seq.append(ChData(I32Type(val), TimeType(), ch_temp))
        return seq

    def get_severity_event(self, severity="DIAGNOSTIC"):
        name = f"apiTester.Severity{severity}"
        temp = self.pipeline.dictionaries.event_name[name]
        event = EventData(tuple(), TimeType(), temp)
        return event

    def get_severity_sequence(self, length, severity="DIAGNOSTIC"):
        return [self.get_severity_event(severity) for _ in range(length)]

    class AssertionFailure(Exception):
        """
        Used to differentiate an AssertionError in test cases that intentionally raise an
        assertion error.
        """

    ######################################################################################
    #   Test Cases
    ######################################################################################
    def test_dummy_pipeline(self):
        length = 15
        event_list = self.get_severity_sequence(length)
        t1 = self.fill_history_async(self.pipeline.enqueue_event, event_list, 0.1)
        print("waiting for queue to fill")
        pred = predicates.greater_than_or_equal_to(length // 2)
        results = self.api.await_event_count(pred)
        assert pred(len(results)), "the correct amount of objects was received"
        t1.join()
        evr_hist = self.api.get_event_test_history()
        item_count = len(evr_hist)
        assert (
            item_count == length
        ), f"Were the correct number of items in the history? ({item_count},{length})"

    def test_find_history_item(self):
        self.fill_history(self.tHistory.data_callback, range(50))

        self.fill_history(self.tHistory.data_callback, range(50))

        pred = predicates.equal_to(25)

        result = self.api.find_history_item(pred, self.tHistory)
        assert result == 25, f"The search should have returned 25, but found {result}"
        result = self.api.find_history_item(pred, self.tHistory, start=50)
        assert result == 25, f"The search should have returned 25, but found {result}"
        result = self.api.find_history_item(pred, self.tHistory, start=80)
        assert (
            result is None
        ), f"The search should have returned None, but found {result}"

    def test_find_history_item_timeout(self):
        pred = predicates.equal_to(25)

        listA = range(50)
        self.fill_history_async(self.tHistory.data_callback, listA, 0.01)
        result = self.api.find_history_item(pred, self.tHistory, timeout=1)
        assert result == 25, f"The search should have returned 25, but found {result}"

        pred = predicates.equal_to(49)
        result = self.api.find_history_item(pred, self.tHistory, timeout=1)
        assert result == 49, f"The search should have returned 49, but found {result}"

        self.tHistory.clear()

        listA = range(50)
        pred = predicates.equal_to(49)
        self.fill_history_async(self.tHistory.data_callback, listA, 0.1)
        result = self.api.find_history_item(pred, self.tHistory, timeout=1)
        assert (
            result is None
        ), f"The search should have returned None, but found {result}"

    def test_find_history_sequence(self):
        sequence = [predicates.equal_to(i) for i in range(30, 40, 2)]

        self.fill_history(self.tHistory.data_callback, range(50))
        results = self.api.find_history_sequence(sequence, self.tHistory)
        assert len(results) == len(
            sequence
        ), f"The search should have found {range(30, 40, 2)}, but returned {results}"
        self.assert_lists_equal(range(30, 40, 2), results)

        results = self.api.find_history_sequence(sequence, self.tHistory, start=34)
        assert len(results) != len(
            sequence
        ), f"The search should have returned an incomplete list, but found {results}"

        self.fill_history(self.tHistory.data_callback, range(50))
        results = self.api.find_history_sequence(sequence, self.tHistory, start=34)
        assert len(results) == len(
            sequence
        ), f"The search should have found {range(30, 40, 2)}, but returned {results}"
        self.assert_lists_equal(range(30, 40, 2), results)

        results = self.api.find_history_sequence(sequence, self.tHistory, start=90)
        assert len(results) != len(
            sequence
        ), f"The search should have returned an incomplete list, but found {results}"

    def test_find_history_sequence_timeout(self):
        sequence = [predicates.equal_to(i) for i in range(30, 40, 2)]

        self.fill_history_async(self.tHistory.data_callback, range(50), 0.01)
        results = self.api.find_history_sequence(sequence, self.tHistory, timeout=1)
        assert results is not None, "The search should have found a sequence"
        self.assert_lists_equal(range(30, 40, 2), results)

        self.fill_history_async(self.tHistory.data_callback, range(50), 0.01)
        results = self.api.find_history_sequence(
            sequence, self.tHistory, start=34, timeout=1
        )
        assert results is not None, "The search should have found a sequence"
        self.assert_lists_equal(range(30, 40, 2), results)

        self.tHistory.clear()

        self.fill_history_async(self.tHistory.data_callback, range(25, 50), 0.1)
        results = self.api.find_history_sequence(
            sequence, self.tHistory, start=90, timeout=1
        )
        assert len(results) != len(
            sequence
        ), f"The search should have returned an incomplete list, but found {results}"

    def test_find_history_count(self):
        count_pred = predicates.greater_than_or_equal_to(10)
        search_pred = predicates.greater_than_or_equal_to(40)

        self.fill_history(self.tHistory.data_callback, range(50))
        results = self.api.find_history_count(count_pred, self.tHistory)
        self.assert_lists_equal(range(50), results)

        results = self.api.find_history_count(count_pred, self.tHistory, search_pred)
        self.assert_lists_equal(range(40, 50), results)

        self.fill_history(self.tHistory.data_callback, range(50, 70))

        results = self.api.find_history_count(count_pred, self.tHistory, search_pred)
        self.assert_lists_equal(range(40, 70), results)

        results = self.api.find_history_count(count_pred, self.tHistory, start=60)
        self.assert_lists_equal(range(60, 70), results)

    def test_find_history_count_timeout(self):
        count_pred = predicates.greater_than_or_equal_to(10)
        search_pred = predicates.greater_than_or_equal_to(40)

        self.fill_history_async(self.tHistory.data_callback, range(50), 0.01)

        results = self.api.find_history_count(count_pred, self.tHistory)
        assert (
            len(results) < 10
        ), f"The search should have returned an incomplete list, but found {results}"

        results = self.api.find_history_count(
            count_pred, self.tHistory, search_pred, timeout=2
        )
        self.assert_lists_equal(range(40, 50), results)

        self.fill_history_async(self.tHistory.data_callback, range(50, 60), 0.01)

        results = self.api.find_history_count(
            count_pred, self.tHistory, search_pred, start=50, timeout=2
        )
        self.assert_lists_equal(range(50, 60), results)

        self.tHistory.clear()

        self.fill_history_async(self.tHistory.data_callback, range(35, 60), 0.1)
        results = self.api.find_history_count(
            count_pred, self.tHistory, search_pred, timeout=1
        )
        assert (
            len(results) < 10
        ), f"The search should have returned an incomplete list, but found {results}"

    def test_get_latest_fsw_time(self):
        ts0 = self.api.get_latest_time()

        time.sleep(0.1)

        ts1 = self.api.get_latest_time()
        assert (
            ts0 is ts1
        ), "The starting timestamp should not have changed if no dataobjects were enqueued"

        count_seq = self.get_counter_sequence(100)
        event_seq = self.get_severity_sequence(100)
        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.02)
        t2 = self.fill_history_async(self.pipeline.enqueue_event, event_seq, 0.02)

        last = ts0
        for i in range(1, 10):
            time.sleep(0.1)
            tsi = self.api.get_latest_time()
            assert tsi > last, f"Iter {i}: {tsi} should be greater than {last}"
            last = tsi

        t1.join()
        t2.join()

        tsn_1 = self.api.get_latest_time()
        assert (
            tsn_1 > last
        ), f"The final timestamp, {tsn_1}, should be greater than {last}."

        time.sleep(0.1)

        tsn_2 = self.api.get_latest_time()
        assert (
            tsn_2 == tsn_1
        ), "The timestamp should not have changed, while no data was streaming."

    def test_clear_histories(self):
        eventHistory = self.api.get_event_test_history()
        channelHistory = self.api.get_telemetry_test_history()
        commandHistory = self.api.get_command_test_history()

        self.api.clear_histories()
        assert eventHistory.size() == 0, "eventHistory should be empty"
        assert channelHistory.size() == 0, "channelHistory should be empty"

        count_seq = self.get_counter_sequence(100)
        event_seq = self.get_severity_sequence(100)
        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.02)
        t2 = self.fill_history_async(self.pipeline.enqueue_event, event_seq, 0.02)
        t1.join()
        t2.join()

        sizeE = eventHistory.size()
        iE = sizeE // 2
        firstE = eventHistory[iE]
        timeE = firstE.get_time()

        sizeC = channelHistory.size()
        iC = next(
            (
                i
                for i in range(channelHistory.size())
                if channelHistory[i].get_time() >= timeE
            ),
            0,
        )
        firstC = channelHistory[iC]

        self.api.clear_histories(timeE)
        msg = f"The event history should have been reduced by {iE} elements"
        assert sizeE - iE == eventHistory.size(), msg
        msg = "The element with the timestamp should be first in the history"
        assert firstE is eventHistory[0], msg
        msg = f"The channel history should have been reduced by {iC} elements"
        assert sizeC - iC == channelHistory.size(), msg
        msg = "The first element in the history should be the first with a valid time"
        assert firstC is channelHistory[0], msg

        args1 = []
        self.api.send_command("apiTester.TEST_CMD_1", args1)
        assert commandHistory.size() > 0, "history size should be greater than 0"
        assert channelHistory.size() > 0, "history size should be greater than 0"
        assert eventHistory.size() > 0, "history size should be greater than 0"

        self.api.clear_histories()

        assert commandHistory.size() == 0, "history size should be 0"
        assert channelHistory.size() == 0, "history size should be 0"
        assert eventHistory.size() == 0, "history size should be 0"

    def test_registering_and_removing_subhistories(self):
        # Verifying that retrieving a subhistory for events behaves as expected
        event_hist = self.api.get_event_test_history()
        self.pipeline.enqueue_event(self.get_severity_event())
        assert event_hist.size() == 1, "There should be one event in the api's history"

        event_subhist = self.api.get_event_subhistory()
        assert event_subhist.size() == 0, "There should be no events in the subhistory"

        self.pipeline.enqueue_event(self.get_severity_event())

        assert event_hist.size() == 2, "There should be two events in the api's history"
        assert event_subhist.size() == 1, "There should be one event in the subhistory"

        assert self.api.remove_event_subhistory(event_subhist), "remove should succeed"

        self.pipeline.enqueue_event(self.get_severity_event())

        assert (
            event_hist.size() == 3
        ), "There should be three events in the api's history"
        assert event_subhist.size() == 1, "There should be one event in the subhistory"

        self.api.clear_histories()

        assert event_hist.size() == 0, "There should be no events in the api's history"
        assert event_subhist.size() == 1, "There should be one event in the subhistory"

        assert not self.api.remove_event_subhistory(
            event_subhist
        ), "should not remove twice"

        # same checks, but for telemetry
        telem_seq = self.get_counter_sequence(3)
        telem_hist = self.api.get_telemetry_test_history()
        self.pipeline.enqueue_telemetry(telem_seq[0])
        assert telem_hist.size() == 1, "There should be one update in the api's history"

        telem_subhist = self.api.get_telemetry_subhistory()
        assert telem_subhist.size() == 0, "There should be no updates in the subhistory"

        self.pipeline.enqueue_telemetry(telem_seq[1])

        assert (
            telem_hist.size() == 2
        ), "There should be two updates in the api's history"
        assert telem_subhist.size() == 1, "There should be one update in the subhistory"

        assert self.api.remove_telemetry_subhistory(
            telem_subhist
        ), "remove should succeed"

        self.pipeline.enqueue_telemetry(telem_seq[2])

        assert (
            telem_hist.size() == 3
        ), "There should be three updates in the api's history"
        assert telem_subhist.size() == 1, "There should be one update in the subhistory"

        self.api.clear_histories()

        assert telem_hist.size() == 0, "There should be no updates in the api's history"
        assert telem_subhist.size() == 1, "There should be one update in the subhistory"

        assert not self.api.remove_telemetry_subhistory(
            telem_subhist
        ), "should not remove twice"

    def test_translate_command_name(self):
        assert self.api.translate_command_name("apiTester.TEST_CMD_1") == 1
        assert self.api.translate_command_name("apiTester.TEST_CMD_2") == 2
        assert self.api.translate_command_name("apiTester.TEST_CMD_3") == 3
        assert self.api.translate_command_name(1) == 1
        assert self.api.translate_command_name(2) == 2
        assert self.api.translate_command_name(3) == 3
        try:
            self.api.translate_command_name("DOES_NOT_EXIST")
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"
        try:
            self.api.translate_command_name(0)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"

    def test_send_command(self):
        args1 = []
        self.api.send_command("apiTester.TEST_CMD_1", args1)
        self.api.send_command(1, args1)
        args2 = ["0x01", "0x02"]
        self.api.send_command("apiTester.TEST_CMD_2", args2)
        self.api.send_command(2, args2)
        args3 = ["test message for the test command"]
        self.api.send_command("apiTester.TEST_CMD_3", args3)
        self.api.send_command(3, args3)
        hist = self.api.get_command_test_history()
        assert hist.size() == 6
        for cmd in hist:
            print(cmd)

    def test_send_and_await_telemetry(self):
        result = self.api.send_and_await_telemetry(
            "apiTester.TEST_CMD_1", channels="CommandCounter"
        )
        assert (
            result is not None
        ), "the search should find the telemetry generated by UTPipeline"

        self.api.clear_histories()

        seq = ["CommandCounter"] + ["Counter"] * 5
        self.fill_history_async(
            self.pipeline.enqueue_telemetry, self.get_counter_sequence(10), 0.01
        )
        results1 = self.api.send_and_await_telemetry(
            "apiTester.TEST_CMD_1", channels=seq
        )
        assert len(results1) == 6, "Should have gotten 6 results out of the await"

        self.fill_history_async(
            self.pipeline.enqueue_telemetry, self.get_counter_sequence(10), 0.01
        )
        results2 = self.api.send_and_await_telemetry(
            "apiTester.TEST_CMD_1", channels=seq
        )
        assert len(results2) == 6, "Should have gotten 6 results out of the await"

        for i in range(6):
            assert results1[i] != results2[i], "These sequences should be unique items"

        self.api.clear_histories()

        seq = ["CommandCounter"] + ["Oscillator"] * 5
        self.fill_history_async(
            self.pipeline.enqueue_telemetry, self.get_oscillator_sequence(10), 0.01
        )
        results = self.api.send_and_await_telemetry(
            "apiTester.TEST_CMD_1", channels=seq
        )
        assert len(results) == 6, "Should have gotten 6 results out of the await"

    def test_send_and_await_event(self):
        result = self.api.send_and_await_event(
            "apiTester.TEST_CMD_1", events="CommandReceived"
        )
        assert (
            result is not None
        ), "the search should have found the CommandReceived Event"

        self.api.clear_histories()

        seq = ["CommandReceived"] + ["SeverityDIAGNOSTIC"] * 5
        self.fill_history_async(
            self.pipeline.enqueue_event, self.get_severity_sequence(10), 0.01
        )
        results1 = self.api.send_and_await_event("apiTester.TEST_CMD_1", events=seq)
        assert len(results1) == 6, "Should have gotten 6 results out of the await"

        self.fill_history_async(
            self.pipeline.enqueue_event, self.get_severity_sequence(10), 0.01
        )
        results2 = self.api.send_and_await_event("apiTester.TEST_CMD_1", events=seq)
        assert len(results2) == 6, "Should have gotten 6 results out of the await"

        for i in range(6):
            assert results1[i] != results2[i], "These sequences should be unique items"

    def test_send_and_assert_telemetry(self):
        self.api.send_and_assert_telemetry(
            "apiTester.TEST_CMD_1", channels="CommandCounter"
        )

        self.api.clear_histories()

        seq = ["CommandCounter"] + ["Counter"] * 5
        self.fill_history_async(
            self.pipeline.enqueue_telemetry, self.get_counter_sequence(10), 0.01
        )
        results1 = self.api.send_and_assert_telemetry(
            "apiTester.TEST_CMD_1", channels=seq, timeout=5
        )

        self.fill_history_async(
            self.pipeline.enqueue_telemetry, self.get_counter_sequence(10), 0.01
        )
        results2 = self.api.send_and_assert_telemetry(
            "apiTester.TEST_CMD_1", channels=seq, timeout=5
        )

        for i in range(6):
            assert results1[i] != results2[i], "These sequences should be unique items"

        self.api.clear_histories()

        seq = ["CommandCounter"] + ["Oscillator"] * 5
        self.fill_history_async(
            self.pipeline.enqueue_telemetry, self.get_oscillator_sequence(10), 0.01
        )
        self.api.send_and_assert_telemetry(
            "apiTester.TEST_CMD_1", channels=seq, timeout=5
        )

    def test_send_and_assert_event(self):
        self.api.send_and_assert_event("apiTester.TEST_CMD_1", events="CommandReceived")

        self.api.clear_histories()

        seq = ["CommandReceived"] + ["SeverityDIAGNOSTIC"] * 5
        self.fill_history_async(
            self.pipeline.enqueue_event, self.get_severity_sequence(10), 0.01
        )
        results1 = self.api.send_and_assert_event(
            "apiTester.TEST_CMD_1", events=seq, timeout=5
        )

        self.fill_history_async(
            self.pipeline.enqueue_event, self.get_severity_sequence(10), 0.01
        )
        results2 = self.api.send_and_assert_event(
            "apiTester.TEST_CMD_1", events=seq, timeout=5
        )

        for i in range(6):
            assert results1[i] != results2[i], "These sequences should be unique items"

    def test_translate_telemetry_name(self):
        assert self.api.translate_telemetry_name("apiTester.CommandCounter") == 1
        assert self.api.translate_telemetry_name("apiTester.Oscillator") == 2
        assert self.api.translate_telemetry_name("apiTester.Counter") == 3
        assert self.api.translate_telemetry_name(1) == 1
        assert self.api.translate_telemetry_name(2) == 2
        assert self.api.translate_telemetry_name(3) == 3
        try:
            self.api.translate_telemetry_name("DOES_NOT_EXIST")
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"
        try:
            self.api.translate_telemetry_name(0)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"

    def test_translate_telemetry_name_search(self):
        assert self.api.translate_telemetry_name("CommandCounter", force_component=False) == [1]
        assert self.api.translate_telemetry_name("Oscillator", force_component=False) == [2]
        assert self.api.translate_telemetry_name("Counter", force_component=False) == [3, 4]
        assert self.api.translate_telemetry_name(1, force_component=False) == 1
        assert self.api.translate_telemetry_name(2, force_component=False) == 2
        assert self.api.translate_telemetry_name(3, force_component=False) == 3
        try:
            self.api.translate_telemetry_name("DOES_NOT_EXIST", force_component=False)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"
        try:
            self.api.translate_telemetry_name(0, force_component=False)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"

    def test_get_telemetry_pred(self):
        pred = predicates.telemetry_predicate()
        result = self.api.get_telemetry_pred(pred)
        assert pred == result, "should return when channel is already telem_pred"

        update = self.get_counter_sequence(1)[0]
        pred = self.api.get_telemetry_pred(update.get_id(), update.get_val())
        assert pred(update), "predicate should return true when fields are specified"

    def test_await_telemetry(self):
        seq = self.get_counter_sequence(20)
        self.fill_history_async(self.pipeline.enqueue_telemetry, seq[:10], 0.01)
        result = self.api.await_telemetry("Counter", 8)
        assert (
            result is not None
        ), f"Await should have found a correct channel update: {result}"

        time.sleep(1)

        self.fill_history_async(self.pipeline.enqueue_telemetry, seq[10:20], 0.01)
        result = self.api.await_telemetry("Counter", 8)
        assert result is None, f"Await should not have found an update: {result}"

        self.api.clear_histories()

        self.fill_history_async(self.pipeline.enqueue_telemetry, seq, 0.1)
        result = self.api.await_telemetry("Counter", 15, timeout=1)
        assert result is None, f"Await should not have found an update: {result}"

    def test_await_telemetry_sequence(self):
        count_seq = self.get_counter_sequence(20)
        sin_seq = self.get_oscillator_sequence(100)

        search_seq = []
        for i in range(15, 20):
            pred = self.api.get_telemetry_pred("Counter", i)
            search_seq.append(pred)

        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        t2 = self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        results = self.api.await_telemetry_sequence(search_seq)

        assert len(results) == len(search_seq), "lists should have the same length"
        for i, item in enumerate(results):
            msg = predicates.get_descriptive_string(item, search_seq[i])
            assert search_seq[i](item), msg

        t1.join()
        t2.join()
        results = self.api.await_telemetry_sequence(search_seq)
        assert len(results) < len(
            search_seq
        ), "repeating the search should not complete"

        self.api.clear_histories()

        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        t2 = self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        results = self.api.await_telemetry_sequence(search_seq, timeout=1)
        assert len(results) < len(
            search_seq
        ), "repeating the search should not complete"

        t1.join()
        t2.join()

    def test_await_telemetry_count(self):
        count_seq = self.get_counter_sequence(20)
        sin_seq = self.get_oscillator_sequence(100)

        pred = predicates.greater_than_or_equal_to(10)
        search_pred = self.api.get_telemetry_pred("Counter", pred)
        count_pred = predicates.within_range(10, 20)

        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        t2 = self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        results = self.api.await_telemetry_count(count_pred, search_pred)
        msg = predicates.get_descriptive_string(len(results), count_pred)
        assert count_pred(len(results)), msg
        t1.join()
        t2.join()

        self.api.clear_histories()

        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        t2 = self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        results = self.api.await_telemetry_count(100)
        assert len(results) == 100, "await count should have found 100 items"
        t1.join()
        t2.join()

        self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.02)
        results = self.api.await_telemetry_count(100, timeout=1)
        assert len(results) < 100, "await count should have found fewer 100 items"

    def test_assert_telemetry(self):
        seq = self.get_counter_sequence(20)
        self.fill_history_async(self.pipeline.enqueue_telemetry, seq[:10], 0.01)
        self.api.assert_telemetry("Counter", 8, timeout=1)

        time.sleep(1)

        self.fill_history_async(self.pipeline.enqueue_telemetry, seq[10:20], 0.01)
        try:
            self.api.assert_telemetry("Counter", 8, start="NOW", timeout=1)
            raise self.AssertionFailure()
        except AssertionError:
            assert True, "api raised the correct error"
        except self.AssertionFailure:
            assert False, "api failed to raise an assertion error"

        self.api.clear_histories()

        self.fill_history_async(self.pipeline.enqueue_telemetry, seq, 0.1)
        try:
            self.api.assert_telemetry("Counter", 15, timeout=1)
            raise self.AssertionFailure()
        except AssertionError:
            assert True, "api raised the correct error"
        except self.AssertionFailure:
            assert False, "api failed to raise an assertion error"

    def test_assert_telemetry_sequence(self):
        count_seq = self.get_counter_sequence(20)
        sin_seq = self.get_oscillator_sequence(100)

        search_seq = []
        for i in range(15, 20):
            pred = self.api.get_telemetry_pred("Counter", i)
            search_seq.append(pred)

        self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        self.api.assert_telemetry_sequence(search_seq, start="NOW", timeout=5)
        self.api.assert_telemetry_sequence(search_seq)

        time.sleep(1)
        try:
            self.api.assert_telemetry_sequence(search_seq, start="NOW", timeout=5)
            raise self.AssertionFailure()
        except AssertionError:
            assert True, "api raised the correct error"
        except self.AssertionFailure:
            assert False, "api failed to raise an assertion error"

        self.api.clear_histories()

        self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.07)
        self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        try:
            self.api.assert_telemetry_sequence(search_seq, start="NOW", timeout=1)
            raise self.AssertionFailure()
        except AssertionError:
            assert True, "api raised the correct error"
        except self.AssertionFailure:
            assert False, "api failed to raise an assertion error"

    def test_assert_telemetry_count(self):
        count_seq = self.get_counter_sequence(20)
        sin_seq = self.get_oscillator_sequence(100)

        pred = predicates.greater_than_or_equal_to(10)
        search_pred = self.api.get_telemetry_pred("Counter", pred)
        count_pred = predicates.within_range(10, 20)

        self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        self.api.assert_telemetry_count(count_pred, search_pred, timeout=2)
        self.api.assert_telemetry_count(count_pred, search_pred)

        self.api.clear_histories()

        t1 = self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        t2 = self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.01)
        self.api.assert_telemetry_count(100, timeout=2)
        t1.join()
        t2.join()
        try:
            self.api.assert_telemetry_count(100)
            raise self.AssertionFailure()
        except AssertionError:
            assert True, "api raised the correct error"
        except self.AssertionFailure:
            assert False, "api failed to raise an assertion error"

        self.api.clear_histories()

        self.fill_history_async(self.pipeline.enqueue_telemetry, count_seq, 0.05)
        self.fill_history_async(self.pipeline.enqueue_telemetry, sin_seq, 0.02)
        try:
            self.api.assert_telemetry_count(100, timeout=1)
            raise self.AssertionFailure()
        except AssertionError:
            assert True, "api raised the correct error"
        except self.AssertionFailure:
            assert False, "api failed to raise an assertion error"

    def test_translate_event_name(self):
        assert self.api.translate_event_name("apiTester.CommandReceived") == 1
        assert self.api.translate_event_name("apiTester.HistorySizeUpdate") == 2
        assert self.api.translate_event_name("apiTester.SeverityCOMMAND") == 3
        assert self.api.translate_event_name("apiTester.SeverityACTIVITY_LO") == 4
        assert self.api.translate_event_name("apiTester.SeverityACTIVITY_HI") == 5
        assert self.api.translate_event_name("apiTester.SeverityWARNING_LO") == 6
        assert self.api.translate_event_name("apiTester.SeverityWARNING_HI") == 7
        assert self.api.translate_event_name("apiTester.SeverityDIAGNOSTIC") == 8
        assert self.api.translate_event_name("apiTester.SeverityFATAL") == 9
        for i in range(1, 10):
            assert self.api.translate_event_name(i) == i

        try:
            self.api.translate_event_name("DOES_NOT_EXIST")
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"
        try:
            self.api.translate_event_name(0)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"

    def test_translate_event_name_search(self):
        assert self.api.translate_event_name("CommandReceived", force_component=False) == [1]
        assert self.api.translate_event_name("HistorySizeUpdate", force_component=False) == [2]
        assert self.api.translate_event_name("SeverityCOMMAND", force_component=False) == [3]
        assert self.api.translate_event_name("SeverityACTIVITY_LO", force_component=False) == [4]
        assert self.api.translate_event_name("SeverityACTIVITY_HI", force_component=False) == [5]
        assert self.api.translate_event_name("SeverityWARNING_LO", force_component=False) == [6]
        assert self.api.translate_event_name("SeverityWARNING_HI", force_component=False) == [7]
        assert self.api.translate_event_name("SeverityDIAGNOSTIC", force_component=False) == [8]
        assert self.api.translate_event_name("SeverityFATAL", force_component=False) == [9, 10]
        for i in range(1, 11):
            assert self.api.translate_event_name(i, force_component=False) == i

        try:
            self.api.translate_event_name("DOES_NOT_EXIST", force_component=False)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"
        try:
            self.api.translate_event_name(0, force_component=False)
            assert False, "the api should have raised a KeyError"
        except KeyError:
            assert True, "the api raised the correct error"

    def test_get_event_pred(self):
        pred = predicates.event_predicate()
        result = self.api.get_event_pred(pred)
        assert pred == result, "should return when channel is already event_pred"

        message = self.get_severity_event("FATAL")
        pred = self.api.get_event_pred(
            message.get_id(), message.get_args(), message.get_severity()
        )
        assert pred(message), "predicate should return true when fields are specified"


"""
    def test_await_event(self):
        raise NotImplementedError("Test Case is not yet implemented")

    def test_await_event_sequence(self):
        raise NotImplementedError("Test Case is not yet implemented")

    def test_await_event_count(self):
        raise NotImplementedError("Test Case is not yet implemented")

    def test_assert_event(self):
        raise NotImplementedError("Test Case is not yet implemented")

    def test_assert_event_sequence(self):
        raise NotImplementedError("Test Case is not yet implemented")

    def test_assert_event_count(self):
        raise NotImplementedError("Test Case is not yet implemented")
"""

if __name__ == "__main__":
    unittest.main()



"""
dictionaries.py:

Helps the standard pipeline wrangle dictionaries by encapsulating the functionality of dictionary loading into a single
class called "Dictionaries".

@author mstarch
"""

from pathlib import Path


# XML Loaders
import fprime_gds.common.loaders.ch_xml_loader
import fprime_gds.common.loaders.cmd_xml_loader
import fprime_gds.common.loaders.event_xml_loader
import fprime_gds.common.loaders.fw_type_json_loader
import fprime_gds.common.loaders.pkt_json_loader
import fprime_gds.common.loaders.pkt_xml_loader

# JSON Loaders
import fprime_gds.common.loaders.ch_json_loader
import fprime_gds.common.loaders.cmd_json_loader
import fprime_gds.common.loaders.event_json_loader

class Dictionaries:
    """
    Dictionaries class to encapsulate the many different dictionaries used in the system. This includes the following
    dictionary types:

    1. Command IDs to Commands
    2. Command names to Commands
    3. Channel IDs to Channels
    4. Channel names to Channels
    5. Event IDs to Events
    6. Event names to Events
    7. Packet IDs to Packets
    """

    def __init__(self):
        """Constructor of the dictionaries object"""
        self._command_id_dict = None
        self._event_id_dict = None
        self._channel_id_dict = None
        self._command_name_dict = None
        self._event_name_dict = None
        self._channel_name_dict = None
        self._packet_dict = None
        self._fw_type_name_dict = None
        self._versions = None
        self._metadata = None
        self._dictionary_path = None
        self._packet_spec_path = None
        self._packet_set_name = None

    def load_dictionaries(self, dictionary, packet_spec, packet_set_name):
        """
        Loads the dictionaries based on the dictionary path supplied. Optional packet_spec is allowed to specify the
        definitions of packets.

        :param dictionary: dictionary path used for loading dictionaries
        :param packet_spec: specification for packets, or None, for packetized telemetry
        :param packet_set_name: name of packet set in case multiple are available
        """
        # Update the "from" values
        self._dictionary_path = dictionary
        self._packet_spec_path = packet_spec
        self._packet_set_name = packet_set_name

        if Path(dictionary).is_file() and ".json" in Path(dictionary).suffixes:
            # Events
            json_event_loader = (
                fprime_gds.common.loaders.event_json_loader.EventJsonLoader(dictionary)
            )
            self._event_name_dict = json_event_loader.get_name_dict(None)
            self._event_id_dict = json_event_loader.get_id_dict(None)
            # Commands
            json_command_loader = (
                fprime_gds.common.loaders.cmd_json_loader.CmdJsonLoader(dictionary)
            )
            self._command_name_dict = json_command_loader.get_name_dict(None)
            self._command_id_dict = json_command_loader.get_id_dict(None)
            # Channels
            json_channel_loader = fprime_gds.common.loaders.ch_json_loader.ChJsonLoader(
                dictionary
            )
            self._channel_name_dict = json_channel_loader.get_name_dict(None)
            self._channel_id_dict = json_channel_loader.get_id_dict(None)
            # Fw Types
            fw_types_loader = fprime_gds.common.loaders.fw_type_json_loader.FwTypeJsonLoader(
                dictionary
            )
            self._fw_type_name_dict = fw_types_loader.get_name_dict(None)
            # Metadata
            self._versions = json_event_loader.get_versions()
            self._metadata = json_event_loader.get_metadata().copy()
            self._metadata["dictionary_type"] = "json"
            # Each loaders should agree on metadata and versions
            assert (
                json_command_loader.get_metadata()
                == json_channel_loader.get_metadata()
                == json_event_loader.get_metadata()
            ), "Metadata mismatch while loading"
            assert (
                json_command_loader.get_versions()
                == json_channel_loader.get_versions()
                == json_event_loader.get_versions()
            ), "Version mismatch while loading"
        # XML dictionaries
        elif Path(dictionary).is_file():
            # Events
            event_loader = fprime_gds.common.loaders.event_xml_loader.EventXmlLoader()
            self._event_id_dict = event_loader.get_id_dict(dictionary)
            self._event_name_dict = event_loader.get_name_dict(dictionary)
            self._versions = event_loader.get_versions()
            # Commands
            command_loader = fprime_gds.common.loaders.cmd_xml_loader.CmdXmlLoader()
            self._command_id_dict = command_loader.get_id_dict(dictionary)
            self._command_name_dict = command_loader.get_name_dict(dictionary)
            assert (
                self._versions == command_loader.get_versions()
            ), "Version mismatch while loading"
            # Channels
            channel_loader = fprime_gds.common.loaders.ch_xml_loader.ChXmlLoader()
            self._channel_id_dict = channel_loader.get_id_dict(dictionary)
            self._channel_name_dict = channel_loader.get_name_dict(dictionary)
            assert (
                self._versions == channel_loader.get_versions()
            ), "Version mismatch while loading"
            # versions are camelCase to match the metadata field of the JSON dictionaries
            self._metadata = {
                "frameworkVersion": self._versions[0],
                "projectVersion": self._versions[1],
                "dictionary_type": "xml",
            }
        else:
            msg = f"[ERROR] Dictionary '{dictionary}' does not exist."
            raise Exception(msg)
        # Check for packet specification
        if packet_spec is not None:
            packet_loader = fprime_gds.common.loaders.pkt_xml_loader.PktXmlLoader()
            self._packet_dict = packet_loader.get_id_dict(
                packet_spec, self._channel_name_dict
            )
        # Otherwise use JSON dictionary to attempt automatic packet loading
        elif self._metadata["dictionary_type"] == "json":
            packet_loader = fprime_gds.common.loaders.pkt_json_loader.PktJsonLoader(dictionary)
            if packet_set_name is None:
                names = packet_loader.get_packet_set_names(None)
                if len(names) == 0:
                    self._packet_dict = None
                    return
                elif len(names) > 1:
                    raise Exception("[ERROR] Multiple packet sets, must set --packet-set-name")
                packet_set_name = names[0]
            self._packet_dict = packet_loader.get_id_dict(
                None, packet_set_name, self._channel_name_dict
            )
        else:
            self._packet_dict = None

    @property
    def command_id(self):
        """Command dictionary by ID"""
        return self._command_id_dict

    @property
    def event_id(self):
        """Event dictionary by ID"""
        return self._event_id_dict

    @property
    def channel_id(self):
        """Channel dictionary by ID"""
        return self._channel_id_dict

    @property
    def command_name(self):
        """Command dictionary by name"""
        return self._command_name_dict

    @property
    def event_name(self):
        """Event dictionary by name"""
        return self._event_name_dict

    @property
    def channel_name(self):
        """Channel dictionary by name"""
        return self._channel_name_dict
    
    @property
    def fw_type_name(self):
        """Fw type name dictionary by name"""
        return self._fw_type_name_dict

    @property
    def project_version(self):
        """Project version in dictionary"""
        return self._versions[1]

    @property
    def framework_version(self):
        """Framework version in dictionary"""
        return self._versions[0]

    @property
    def metadata(self):
        """Dictionary metadata.

        Note: framework_version and project_version are also available as separate properties
        for legacy reasons. New code should use the metadata property."""
        return self._metadata

    @property
    def dictionary_path(self):
        """ Dictionary Path """
        return self._dictionary_path

    @property
    def packet_spec_path(self):
        """ Dictionary Path """
        return self._packet_spec_path
    
    @property
    def packet_set_name(self):
        """ Dictionary Path """
        return self._packet_set_name

    @property
    def packet(self):
        """Packet dictionary"""
        return self._packet_dict



"""
@brief Base class for all Loaders. Defines the Loader interface

Loaders are responsible for reading dictionaries autocoded during the fprime
build process and converting them to python dictionaries with ids or names as
keys and data_template objects as values.

Additionally, because loading dictionaries is often time consuming, this loader
saves the python dictionaries with the path that produced them so that when
a dictionary from a previously used path is asked for, the dictionary can be
returned immediately.

The base class will return empty dictionaries for any path.

@date Created July 3, 2018
@author R. Joseph Paetz

@bug No known bugs
"""


class DictLoader:
    """
    Base class for all loader classes. Defines the loader interface

    Example:
        l = DictLoader  # Or any other derived loader class
        id_dict = l.get_id_dict(path)       # May take a second or two
        name_dict = l.get_name_dict(path)   # Returns immediately
    """

    def __init__(self):
        """
        Constructor

        The base class constructor sets up the machinery so that id and name
        dictionaries for a given path can be saved and thus not have to be
        remade if they are asked for again.

        Returns:
            An initialized loader object
        """
        self.saved_dicts = {}
        self.versions = ("unknown", "unknown")

    def get_id_dict(self, path):
        """
        Returns the python dictionary keyed by ids for the given path

        This function will return the same dictionary originally computed for
        the given path or will construct new dictionaries if the path has never
        been passed to the get_id_dict or the get_name_dict functions.

        This function should not be overwritten by inheriting classes

        Args:
            path: Path to the file system dictionary to convert to a python dict

        Returns:
            The id dictionary associated with the given path
        """
        if path in self.saved_dicts:
            (id_dict, name_dict) = self.saved_dicts[path]
        else:
            (id_dict, name_dict, self.versions) = self.construct_dicts(path)
            self.saved_dicts[path] = (id_dict, name_dict)

        return id_dict

    def get_name_dict(self, path):
        """
        Returns the python dictionary keyed by names for the given path

        This function will return the same dictionary originally computed for
        the given path or will construct new dictionaries if the path has never
        been passed to the get_id_dict or the get_name_dict functions.

        This function should only be overwritten by inheriting classes if
        more arguments are needed to construct the dictionaries.

        Args:
            path: Path to the file system dictionary to convert to a python dict

        Returns:
            The name dictionary associated with the given path
        """
        if path in self.saved_dicts:
            (id_dict, name_dict) = self.saved_dicts[path]
        else:
            (id_dict, name_dict, self.versions) = self.construct_dicts(path)
            self.saved_dicts[path] = (id_dict, name_dict)

        return name_dict

    def get_versions(self):
        """Get version tuple"""
        return self.versions

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name.

        This function should only be overwritten by inheriting classes if
        more arguments are needed to construct the dictionaries.

        Args:
            path: Path to the file system dictionary to convert to a python dict

        Returns:
            A tuple with two dictionaries: (id_dict, name_dict). They should
            have keys of the id and name fields respectively and the values
            for both should be data_template classes. This base class only
            returns empty dictionaries.
        """
        return {}, {}, ("unknown", "unknown")



"""
framing.py:

This module defines all the framing constructs used when adapting the F prime ground system for use with communications
or elsewhere in the system. There is an abstract base class called FramerDeframe, which defines the standard pattern of
'frame', 'deframe', and the helper 'deframe_all'. This allows users to frame and deframe packets using base classes
that implement this pattern. The current list of implementation classes are:

1. FpFramerDeframer: a class used to write the now-standard F prime ground packet format
2. TcpServerFramerDeframer: a non-symmetric framer/deframer for use interacting with the Tcp Server packet format

@author lestarch
"""

from __future__ import annotations
import abc
import copy
import struct
import sys
from typing import Type

from .checksum import calculate_checksum, CHECKSUM_MAPPING
from fprime_gds.plugin.definitions import (
    gds_plugin_implementation,
    gds_plugin_specification,
)


class FramerDeframer(abc.ABC):
    """
    Abstract base class of the Framer/Deframer variety. Framers and Deframers have to define two methods, one for
    framing a set of bytes and one for deframing a set of bytes into packets.
    """

    @abc.abstractmethod
    def frame(self, data: bytes) -> bytes:
        """
        Frames outgoing data in the specified format. Expects incoming raw bytes to frame, and adds on the needed header
        and footer bytes. This new array of bytes is returned from the method.

        :param data: bytes to frame
        :return: array of raw bytes representing a framed packet. Should be ready for uplink.
        """

    @abc.abstractmethod
    def deframe(
        self, data: bytes, no_copy=False
    ) -> tuple[(bytes | None), bytes, bytes]:
        """
        Deframes the incoming data from the specified format.
        Produces:
        - One packet, or None if no packet found
        - leftover bytes (not consumed yet)
        - Discarded data (consumed and determined not to be valid)

        Users wanting all packets to be deframed should call "deframe_all". If no full packet is available, this method
        returns None. Expects incoming raw bytes to deframe, and returns a deframed packet or None, the leftover
        bytes that were unused, and any bytes discarded from the existing data stream. Will search and discard data up
        until a start token is found. Note: data will be consumed up to the first start token found.

        :param data: framed data bytes
        :param no_copy: (optional) will prevent extra copy if True, but "data" input will be destroyed.
        :return: (packet as bytes or None, leftover bytes, any discarded data)
        """

    def deframe_all(self, data: bytes, no_copy: bool):
        """
        Deframes all available packets found in a single set of bytes by calling deframe until a None packet is
        retrieved. This list of packets, and the remaining bytes are returned

        :param data: framed data bytes
        :param no_copy: (optional) will prevent extra copy if True, but "data" input will be destroyed.
        :return: list of packets, remaining data, discarded/unframed/garbage data
        """
        packets = []
        if not no_copy:
            data = copy.copy(data)
        discarded_aggregate = b""
        while True:
            # Deframe and return only on None
            (deframed, data, discarded) = self.deframe(data, no_copy=True)
            discarded_aggregate += discarded
            if deframed is None:  # No more packets available, return aggregate
                return packets, data, discarded_aggregate
            packets.append(deframed)

    @classmethod
    @gds_plugin_specification
    def register_framing_plugin(cls) -> Type["FramerDeframer"]:
        """Register a plugin to provide framing capabilities

        Plugin hook for registering a plugin that supplies a FramerDeframer implementation. Implementors of this hook must
        return a non-abstract subclass of FramerDeframer. This class will be provided as a framing implementation option
        that users may select via command line arguments.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            FramerDeframer subclass
        """
        raise NotImplementedError()


class FpFramerDeframer(FramerDeframer):
    """
    Framing object used to read/write in the standard F prime format. This format is compatible with the standard ground
    system. It contains the following format for data:

    | token: start word |
    | token: data length |
    | bytes: F prime packet bytes (Fw::Comm, Fw::FilePacket, etc.) |
    | token: checksum |

    Where a token is a big-endian integer of TOKEN_SIZE bytes in length (see below)
    """

    # Size of an F prime framing token, and the type based on that size
    TOKEN_SIZE = 4
    # Total size of header data based on token size
    HEADER_SIZE = TOKEN_SIZE * 2
    # Size of checksum value, and the hardcoded value before CRC32 is available
    CHECKSUM_SIZE = 4
    MAXIMUM_DATA_SIZE = 4096

    # Filled by set_constants()
    TOKEN_TYPE = None
    HEADER_FORMAT = None
    START_TOKEN = None

    def __init__(self, checksum_type="crc32"):
        """Sets constants on construction."""
        # Setup the constants as soon as possible.
        FpFramerDeframer.set_constants()
        self.checksum = checksum_type

    @classmethod
    def set_constants(cls):
        """
        Setup the constants for the various token sizes. This will ensure that the system can read the tokens properly.
        This can be changed to make the framing more efficient.
        """
        if FpFramerDeframer.TOKEN_SIZE == 4:
            FpFramerDeframer.TOKEN_TYPE = "I"
            FpFramerDeframer.START_TOKEN = 0xDEADBEEF
        elif FpFramerDeframer.TOKEN_SIZE == 2:
            FpFramerDeframer.TOKEN_TYPE = "H"
            FpFramerDeframer.START_TOKEN = 0xBEEF
        elif FpFramerDeframer.TOKEN_SIZE == 1:
            FpFramerDeframer.TOKEN_TYPE = "B"
            FpFramerDeframer.START_TOKEN = 0xEF
        else:
            msg = f"Invalid TOKEN_SIZE of {FpFramerDeframer.TOKEN_SIZE}"
            raise ValueError(msg)
        FpFramerDeframer.HEADER_FORMAT = ">" + (FpFramerDeframer.TOKEN_TYPE * 2)

    def frame(self, data):
        """
        Frames outgoing data in the F prime standard format. Expects incoming raw bytes to frame, and adds on the
        needed framing tokens to the front and end of the bytes.

        :param data: bytes to frame
        :return: array of raw bytes representing a framed packet. Should be ready for uplink.
        """
        framed = struct.pack(
            FpFramerDeframer.HEADER_FORMAT, FpFramerDeframer.START_TOKEN, len(data)
        )
        framed += data
        framed += struct.pack(">I", calculate_checksum(framed, self.checksum))
        return framed

    def deframe(self, data, no_copy=False):
        """
        Deframes the incoming data from the F prime standard format. Produces exactly one packet, and leftover bytes.
        Users wanting all packets to be deframed should call "deframe_all". If no full packet is available, this method
        returns None. Expects incoming raw bytes to deframe, and returns a deframed packet or None, and the leftover
        bytes that were unused. Will search and discard data up until a start token is found. Note: data will be
        consumed up to the first start token found.

        :param data: framed data bytes
        :param no_copy: (optional) will prevent extra copy if True, but "data" input will be destroyed.
        :return: (packet as array of bytes or None, leftover bytes)
        """
        discarded = b""
        if not no_copy:
            data = copy.copy(data)
        # Continue until there is not enough data for the header, or until a packet is found (return)
        while len(data) >= FpFramerDeframer.HEADER_SIZE:
            # Read header information including start token and size and check if we have enough for the total size
            start, data_size = struct.unpack_from(FpFramerDeframer.HEADER_FORMAT, data)
            total_size = (
                FpFramerDeframer.HEADER_SIZE
                + data_size
                + FpFramerDeframer.CHECKSUM_SIZE
            )
            # Invalid frame, rotate away a Byte and keep processing
            if (
                start != FpFramerDeframer.START_TOKEN
                or data_size >= FpFramerDeframer.MAXIMUM_DATA_SIZE
            ):
                discarded += data[0:1]
                data = data[1:]
                continue
            # If the pool is large enough to read the whole frame, then read it
            if len(data) >= total_size:
                deframed, check = struct.unpack_from(
                    f">{data_size}sI", data, FpFramerDeframer.HEADER_SIZE
                )
                # If the checksum is valid, return the packet. Otherwise continue to rotate
                if check == calculate_checksum(
                    data[: data_size + FpFramerDeframer.HEADER_SIZE], self.checksum
                ):
                    data = data[total_size:]
                    return deframed, data, discarded
                print(
                    "[WARNING] Checksum validation failed.",
                    file=sys.stderr,
                )
                # Bad checksum, rotate 1 and keep looking for non-garbage
                discarded += data[0:1]
                data = data[1:]
                continue
            # Case of not enough data for a full packet, return hoping for more later
            return None, data, discarded
        return None, data, discarded

    @classmethod
    def get_name(cls):
        """Get the name of this plugin"""
        return "fprime"

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """Register a bad plugin"""
        return cls


class TcpServerFramerDeframer(FramerDeframer):
    """
    Framing object used to read/write in the Tcp Server format. This format is compatible with the older ground
    system. It contains the following formats for data:

    Uplink (command) data:
    | string: 'ZZZZ' |
    | bytes: F prime packet bytes (Fw::Comm, Fw::FilePacket, etc.) |

    Downlink (ground-system) data:
    | string: 'A5A5 GUI ' |
    | U32: length of data |
    | bytes: F prime packet bytes (Fw::Comm, Fw::FilePacket, etc.) |

    WARNING: this is a non-symmetric framer/deframer as the Tcp Server's protocol isn't symmetric. Therefore, it should
    never be used to deframe its own framed data.
    """

    START_STRING = "ZZZZ"

    def frame(self, data):
        """
        Frames outgoing data in the Tcp server outgoing format. Expects incoming raw bytes to frame, and adds on the
        needed framing tokens to the front and end of the bytes.

        :param data: bytes to frame
        :return: array of raw bytes representing a framed packet. Should be ready for uplink.
        """
        return b"A5A5 GUI %s%s" % (struct.pack(">I", len(data)), data)

    def deframe(self, data, no_copy=False):
        """
        Deframes the incoming data from the F prime standard format. Produces exactly one packet, and leftover bytes.
        Users wanting all packets to be deframed should call "deframe_all". If no full packet is available, this method
        returns None. Expects incoming raw bytes to deframe, and returns a deframed packet or None, and the leftover
        bytes that were unused. Will search and discard data up until a start token is found. Note: data will be
        consumed up to the first start token found.

        :param data: framed data bytes
        :param no_copy: (optional) will prevent extra copy if True, but "data" input will be destroyed.
        :return: (packet as array of bytes or None, leftover bytes)
        """
        if not no_copy:
            data = copy.copy(data)
        # Shift over to ZZZZ
        while len(data) > 4 and data[:4] != b"ZZZZ":
            data = data[1:]
        # Break out of data when not enough
        if len(data) < 8:
            return None, data, b""
        # Read the length and break if not enough data
        (data_len,) = struct.unpack_from(">I", data, 4)
        if len(data) < data_len + 8:
            return None, data, b""
        packet = data[8 : data_len + 8]
        data = data[data_len + 8 :]
        return packet, data, b""
